@@ -1,5 +1,5 @@
 /*
- * $Id: ft_conn.c,v 1.55 2004/05/05 04:42:42 jasta Exp $
+ * $Id: ft_conn.c,v 1.57 2004/11/14 00:34:14 hexwab Exp $
  *
  * Copyright (C) 2001-2003 giFT project (gift.sourceforge.net)
  *
@@ -414,6 +414,9 @@ int ft_conn_maintain (void *udata)
 	if (timer_cnt & 1)
 		ft_node_cache_update ();
 
+	if (timer_cnt & 1)
+		ft_session_tidy_streams ();
+
 	/* gather stats, either from our index nodes if we are a search node or
 	 * our parent if we are a child */
 	n = ft_netorg_foreach (FT_NODE_INDEX, FT_NODE_CONNECTED, 10,
@@ -563,7 +566,7 @@ static int get_fd_limit (void)
 	 * course */
 	if (limit < 4096)
 	{
-		rlim.rlim_cur = MAX (rlim.rlim_max, 4096);
+		rlim.rlim_cur = MIN (rlim.rlim_max, 4096);
 
 		if ((ret = setrlimit (RLIMIT_OFILE, &rlim)) != 0)
 		{
