@@ -1,5 +1,5 @@
 /*
- * $Id: ft_search_db.c,v 1.138 2004/10/11 22:13:21 hexwab Exp $
+ * $Id: ft_search_db.c,v 1.145 2005/04/18 21:45:55 jasta Exp $
  *
  * Copyright (C) 2001-2003 giFT project (gift.sourceforge.net)
  *
@@ -67,6 +67,9 @@ static BOOL db_initialized = FALSE;
  */
 /* #define PARANOID */
 
+/* be verbose about single removals in an attempt to track down greycat's bug */
+/* #define DEBUG_REMOVE */
+
 /*
  * Sync all databases after changes have been made to ease debugging.  This
  * greatly hurts performance and should not be enabled for every day usage.
@@ -128,6 +131,15 @@ static BOOL db_initialized = FALSE;
 
 /*****************************************************************************/
 
+#define DB_BITS 12
+#define ID_BITS 20
+
+/* this must be larger than the search/children config key! */
+#define MAX_CHILDREN   (1 << DB_BITS) 
+
+/* local_child is always the first db created, so it gets id zero */
+#define LOCAL_CHILD    0
+
 /*
  * All of the globally opened environment and database handlers go here.
  * Keep in mind that the actual share record data goes in a separate database
@@ -141,6 +153,11 @@ static DB     *db_token_idx    = NULL;
 static DB     *db_share_data   = NULL;
 FTSearchDB    *local_child     = NULL;
 
+/* NOTE: this is not actually a globally-unique ID, merely unique
+ * per-child.  But to save tracking which IDs each child has used we
+ * just keep a single counter and increment and retry when insertion
+ * fails.
+ */
 static uint32_t global_id = 0;
 
 #ifndef SEARCH_DB_BENCHMARK
@@ -175,8 +192,15 @@ struct tokenidx_key
 
 struct tokenidx_data
 {
+#if 0
 	FTSearchDB   *sdb;                 /* raw pointer into the database */
 	uint32_t      id;
+#else
+	/* FIXME: evil limits here! 4096 total children, ~1 million
+	 * total files (global) */
+	uint32_t      db_id:DB_BITS;
+	uint32_t      id:ID_BITS;
+#endif
 	unsigned char order;
 	unsigned char count;
 }
@@ -282,6 +306,66 @@ static BOOL db_abort (FTSearchDB *sdb);
 
 /*****************************************************************************/
 
+static FTSearchDB *child_index[MAX_CHILDREN];
+static int last_child_id, child_count;
+
+/* returns a unique child ID */
+static int child_new (FTSearchDB *sdb)
+{
+	int id = last_child_id;
+
+	assert (sdb);
+
+	/* FIXME: the count includes children that have disconnected
+	 * and are pending removal, so this can trigger in unfortunate
+	 * situations where the removal queue is long despite the
+	 * child count being clamped below the limit.  (Yes, this has
+	 * happened to me.)
+	*/
+	child_count++;
+	assert (child_count < MAX_CHILDREN);
+
+	while (child_index[id])
+		id++;
+
+	child_index[id] = sdb;
+
+	return id;
+}
+
+static void child_free (int id)
+{
+	assert (child_index[id]);
+
+	child_count--;
+
+	child_index[id] = NULL;
+}
+
+static FTSearchDB *child_lookup (int id)
+{
+	FTSearchDB *sdb;
+
+	assert (id >= 0 && id < MAX_CHILDREN);
+
+	sdb = child_index[id];
+
+	assert (sdb);
+
+	return sdb;
+}
+
+static void init_children (void)
+{
+	int i;
+	for (i=0; i < MAX_CHILDREN; i++)
+		child_index[i] = NULL;
+
+	last_child_id = child_count = 0;
+}
+
+/*****************************************************************************/
+
 static FTSearchDB *search_db_new (FTNode *node)
 {
 	FTSearchDB *sdb;
@@ -303,6 +387,11 @@ static FTSearchDB *search_db_new (FTNode *node)
 	
 	sdb->node = node;
 
+	sdb->id = child_new (sdb);
+
+	FT->DBGFN (FT, "db_new: %s (%p) has id %d (0x%x)", ft_node_fmt (node),
+		   sdb, sdb->id, sdb->id);
+
 	return sdb;
 }
 
@@ -327,6 +416,10 @@ static void search_db_free (FTSearchDB *sdb)
 	assert (sdb->old_filter == NULL);
 #endif
 
+	child_free (sdb->id);
+
+	FT->DBGFN (FT, "db_free: freed %p with id %d (0x%x)", sdb, sdb->id, sdb->id);
+
 	free (sdb->share_idx_name);
 	free (sdb);
 }
@@ -674,8 +767,8 @@ static DB_COMPAREFN (compare_sdb)
 
 	if (ret)
 		return ret;
-
-	return a_rec->id - b_rec->id;
+	
+	return memcmp (&a_rec->id, &b_rec->id, sizeof (a_rec->id));
 }
 
 static DB *db_md5idx (void)
@@ -689,29 +782,26 @@ static DB *db_md5idx (void)
 
 static DB_COMPAREFN (compare_md5)
 {
-	static struct tokenidx_data *a_rec;
-	static struct tokenidx_data *b_rec;
+	static struct tokenidx_data a_rec, b_rec;
 	int ret;
 
-	a_rec = a->data;
-	b_rec = b->data;
-
-	assert (a->size == sizeof (*a_rec));
+	assert (a->size == sizeof (a_rec));
 	assert (b->size == a->size);
 
+	/* We need to copy the structs because libdb doesn't guarantee
+	 * alignedness.  Neither can we use memcmp (despite not caring
+	 * about the relative ordering, only the uniqueness), as the
+	 * fields aren't even byte-aligned.
+	 */
+	a_rec = *(struct tokenidx_data *)a->data;
+	b_rec = *(struct tokenidx_data *)b->data;
+
 	/* first compare id */
-	if ((ret = a_rec->id - b_rec->id))
+	if ((ret = a_rec.id - b_rec.id))
 		return ret;
 
-	/*
-	 * We still need the ability to insert two md5s from different hosts, so
-	 * we have to compare the host here in order to allow the insert to
-	 * succeed.
-	 *
-	 * BUG: The sort order is based on hte pointers value, and this is
-	 * probably not what we want.
-	 */
-	return memcmp (&a_rec->sdb, &b_rec->sdb, sizeof (a_rec->sdb));
+	/* then compare host */
+	return a_rec.db_id - b_rec.db_id;
 
 	/* the rest of the record is ignored, so that we don't have to
 	 * worry about filling it in and it'll still match */
@@ -1030,6 +1120,7 @@ static uint32_t db_insert_sharedata (FTSearchDB *sdb, Hash *hash,
 	DBT data;
 	int ret;
 	uint32_t id;
+	int looped = 0;
 
 	/* access the appropriate database, opening if necessary */
 	if (!(dbp = db_sharedata()))
@@ -1067,11 +1158,19 @@ static uint32_t db_insert_sharedata (FTSearchDB *sdb, Hash *hash,
 
 	/* choose a new id, retrying until it succeeds */
 	do {
-		
 		/* skip zero */
-		do {
-			id = global_id++;
-		} while (!id);
+		id = global_id = (global_id + 1) & ((1<<ID_BITS)-1);
+		if (!id)
+		{
+			id = ++global_id;
+
+			/* abort if we have too many shares */
+			if (looped++)
+			{
+				db_abort (sdb);
+				return 0;
+			}
+		}
 
 		keyrec.sdb = sdb;
 		keyrec.id  = id;
@@ -1142,7 +1241,7 @@ static BOOL db_insert_tokenidx (FTSearchDB *sdb, Hash *hash, struct tokenized *t
 	if (dbp->cursor (dbp, NULL, &dbcp, 0))
 		return FALSE;
 
-	datarec.sdb = sdb;
+	datarec.db_id = sdb->id;
 	datarec.id  = id;
 
 	memset (&key, 0, sizeof (key));
@@ -1253,7 +1352,7 @@ static BOOL db_remove_tokenidx_token (DBC *dbcp, FTSearchDB *sdb,
 
 	keyrec.token = token;
 
-	datarec.sdb = sdb;
+	datarec.db_id = sdb->id;
 	datarec.id  = id;
 
 	memset (&key, 0, sizeof (key));
@@ -1500,6 +1599,7 @@ static void check_tokens (FTSearchDB *sdb)
 	struct tokenidx_key *keyrec;
 	struct tokenidx_data *datarec;
 	int  count = 0;
+	int  id = sdb->id;
 
 	if (!(dbp = db_tokenidx()))
 		db_abort (sdb);
@@ -1523,7 +1623,8 @@ static void check_tokens (FTSearchDB *sdb)
 		keyrec = key.data;
 		datarec = data.data;
 		
-		if (datarec->sdb == sdb)
+		if (datarec->db_id == id ||
+		    !child_index[datarec->db_id])
 			db_abort (sdb);
 
 		count++;
@@ -2331,22 +2432,25 @@ static List *token_gather_cursors (DB *dbp, struct tokenized *query,
 static BOOL is_stale_db (DBT *data)
 {
 	struct tokenidx_data *datarec;
+	FTSearchDB *sdb;
 
 	assert (data->size == sizeof (*datarec));
 	datarec = data->data;
 
-	if (datarec->sdb == local_child)
+	if (datarec->db_id == LOCAL_CHILD)
 		return FALSE;
 
+	sdb = child_lookup (datarec->db_id);
+
 #ifndef SEARCH_DB_BENCHMARK
-	if (datarec->sdb->node == NULL)
+	if (sdb->node == NULL)
 		assert (remove_active == TRUE);
 #else
-	assert (datarec->sdb->node != NULL);
+	assert (sdb->node != NULL);
 #endif
 
 	/* if node is NULL, stale is TRUE */
-	return BOOL_EXPR (datarec->sdb->node == NULL);
+	return BOOL_EXPR (sdb->node == NULL);
 }
 
 static BOOL look_for (struct cursor_stream *s, DBT *data_cmp)
@@ -2649,15 +2753,16 @@ static int match_tokens (Array **results, List **qt, int max_results,
 			struct tokenidx_data *datarec = data.data;
 			uint8_t *share_order;
 			Share *share;
-			
+			FTSearchDB *sdb = child_lookup (datarec->db_id);
+
 			/* grab the Share and order list */
-			if (!(share = db_get_share (datarec->sdb, datarec->id,
-						     &share_order)))
+			if (!(share = db_get_share (sdb, datarec->id, &share_order)))
 			{
-				db_abort (datarec->sdb);
 				FT->DBGFN (FT, "%s: unable to lookup id %d",
-					   ft_node_fmt (datarec->sdb->node),
+					   ft_node_fmt (sdb->node),
 					   datarec->id);
+				db_abort (sdb);
+
 				continue;
 			}
 
@@ -2762,6 +2867,8 @@ BOOL ft_search_db_init (const char *envpath, unsigned long cachesize)
 		env_search_path = NULL;
 	}
 
+	init_children ();
+
 	local_child = search_db_new (NULL);
 #ifdef BLOOM
 	local_child->filter = md5_filter.filter;
@@ -2778,17 +2885,41 @@ void ft_search_db_destroy (void)
 		return;
 
 #ifdef USE_LIBDB
-	if (env_search_path)
+	/* Clean up all the individual dbs.  Even though the db
+	 * environment and all it contains is about to be destroyed,
+	 * we need to delete everything to make sure it isn't synced
+	 * to disk (and to clean up all the non-db structures that
+	 * would otherwise leak).
+	 */
 	{
-		if (env_search)
-			db_destroy (env_search_path);
+		int i;
+		
+		/* local_child is included in this list */
+		for (i=0; i < MAX_CHILDREN; i++)
+		{
+			FTSearchDB *sdb = child_index[i];
+
+			if (sdb)
+			{
+				db_close (sdb, TRUE);
+				search_db_free (sdb);
+			}
+		}
+
+		/* FIXME: de-ickify this mess */
+		close_db (db_md5_idx, "md5.index", NULL, TRUE);
+		close_db (db_token_idx, "tokens.index", NULL, TRUE);
+		close_db (db_share_data, "share.data", NULL, TRUE);
+	}
+
+	assert (env_search_path);
+	{
+		db_destroy (env_search_path);
 
 		free (env_search_path);
 		env_search_path = NULL;
 	}
 
-	search_db_free (local_child);
-	local_child = NULL;
 #endif /* USE_LIBDB */
 
 	db_initialized = FALSE;
@@ -3679,6 +3810,9 @@ BOOL test_suite_search_db (Protocol *p)
 #else
 	int ret;
 
+	/* FIXME: should be later */
+	ft_search_db_init ("benchtemp", 50<<20);
+
 	if ((ret = load_test_data ("test.data")) <= 0)
 	{
 		FT->err (FT, "test.data: no test data found");
@@ -3691,7 +3825,6 @@ BOOL test_suite_search_db (Protocol *p)
 		return FALSE;
 	}
 
-	ft_search_db_init ("benchtemp", 50<<20);
 	test_benchmarks ();
 
 	free_test_data ();
