@@ -1,5 +1,5 @@
 /*
- * $Id: ft_transfer.c,v 1.13 2003/12/23 04:12:21 jasta Exp $
+ * $Id: ft_transfer.c,v 1.16 2005/01/02 21:44:40 mkern Exp $
  *
  * Copyright (C) 2001-2003 giFT project (gift.sourceforge.net)
  *
@@ -107,7 +107,11 @@ static BOOL decompose_source (struct ft_source *src, char *url)
 
 	/* move beyond the OpenFT:// prefix */
 	proto = string_sep (&url, "://");
-	assert (strcmp (proto, "OpenFT") == 0);
+
+	/* this can happen if the URL is missing the slashes (the
+	 * proto name and colon are guaranteed) */
+	if (strcmp (proto, "OpenFT") != 0)
+		return FALSE;
 
 	/* divide the string in to two parts, one pointing to
 	 * "ip:http_port" and the other to "requested/file?md5=..." */
